#!/usr/bin/env python
#
# Plot the matrices computed and used by permute.py

import docx
import matplotlib
matplotlib.use('Agg')
import sys
import numpy as np
import pylab
import matplotlib.lines as mpllines
import cpa
import cpa.util
from cpa.profiling.profiles import Profiles
from cpa.profiling.cache import Cache, DummyNormalization
from cpa.profiling.normalization import StdNormalization
import lib
from util import cached

CACHE_DIR = '../../cache/'

cpa.properties.LoadFile("supplement.properties")

@cached("../results/HEAD/cv_plate_raw_features.pickle")
def get_raw_plate_profiles():
    cache = Cache(CACHE_DIR)
    dbres = cpa.db.execute("SELECT DISTINCT ImageNumber, Image_Metadata_PlateID FROM %s WHERE Image_Metadata_ASSAY_WELL_ROLE = 'mock'" % (cpa.properties.image_table))
    images_by_plate = {}
    for image_number, plate in dbres:
        plate = int(plate)
        image_key = (int(image_number),)
        images_by_plate.setdefault(plate, []).append(image_key)

    vector_by_plate = {}
    for plate, images in images_by_plate.items():
        data = cache.load(images, normalization=DummyNormalization)[0]
        vector_by_plate[plate] = data.mean(0)

    return np.array(vector_by_plate.values())

@cached("../results/HEAD/cv_wellpos_raw_features.pickle")
def get_raw_wellpos_profiles():
    cache = Cache('../../cache/')
    by_wellpos = lib.get_mock_images_per_wellpos()
    profiles = []
    for wellpos, by_plate in by_wellpos.items():
        per_plate_profiles = []
        for plate, images in by_plate.items():
            data = cache.load(images, normalization=DummyNormalization)[0]
            per_plate_profiles.append(data)
        profiles.append(np.vstack(per_plate_profiles).mean(0))
    return np.array(profiles)

def calculate_cv(profiles):
    return np.absolute(profiles.std(0) / profiles.mean(0))

def histogram_cv(profiles, label):
    pylab.figure()
    h, e = np.histogram(calculate_cv(profiles), 50)
    pylab.bar(0.5 * e[:-1] + 0.5 * e[1:], h, e[1] - e[0])
    pylab.xlim(0, 3.5)
    pylab.xlabel('Coefficient of variation (absolute)')
    pylab.ylabel('Number of features')
    pylab.savefig('../results/HEAD/cv_histogram_%s.pdf' % label)
    pylab.close()

def ranking_docx(profiles, label):
    cv = calculate_cv(profiles)
    indices = np.argsort(cv)[::-1]
    cache = Cache(CACHE_DIR)
    normalization = DummyNormalization(cache)
    colnames = normalization.colnames
    table = [['Rank', 'CV', 'Feature name']]
    for rank0, i in enumerate(indices):
        table.append(['%d' % (rank0 + 1), '%f' % cv[i], colnames[i]])
    document = docx.newdocument()
    docbody = document.xpath('/w:document/w:body', namespaces=docx.nsprefixes)[0]
    docbody.append(docx.paragraph(label, style='b'))
    docbody.append(docx.table(table))
    docx.savedocx(document,
                  docx.coreproperties(title='Feature table', subject='',
                                      creator='Vebjorn Ljosa',
                                      keywords=[]),
                  docx.appproperties(), docx.contenttypes(),
                  docx.websettings(),
                  docx.wordrelationships(docx.relationshiplist()),
                  '../results/HEAD/cv_ranking_%s.docx' % label)

if __name__ == '__main__':
    plate_profiles = get_raw_plate_profiles()
    wellpos_profiles = get_raw_wellpos_profiles()

    histogram_cv(plate_profiles, "plate")
    histogram_cv(wellpos_profiles, "wellpos")

    ranking_docx(plate_profiles, "plate")
    ranking_docx(wellpos_profiles, "wellpos")

