#!/usr/bin/env
#
# Read profiles and annotations. Exclude profiles for compounds that
# are inactive or do not have annotations.

import sys
import os
import progressbar
import numpy as np
from scipy.spatial.distance import cdist
import cpa
import cpa.util
from cpa.profiling.profiles import Profiles
from annotations import read_annotations, slim_annotations
from common import *

# Exclude inactive profiles
def compute_active_mask(profiles, dmso_well_profiles, percentile):
    dmso_profile = dmso_well_profiles.data.mean(0)
    dmso_distances = cdist([dmso_profile], dmso_well_profiles.data, 'euclidean')[0,:]
    compound_distances = cdist([dmso_profile], profiles.data, 'euclidean')[0,:]
    cutoff = sorted(dmso_distances)[int(len(dmso_distances) * percentile)]
    active_mask = compound_distances >= cutoff
    print '%d of %d compounds (%.0f%%) are active (dist. from DMSO above %.0fth percentile)' % (
        active_mask.sum(), len(active_mask), 
        100.0 * active_mask.sum() / len(active_mask), percentile * 100.0)
    return active_mask

def load_profiles(percentile):
    profiles = Profiles.load('../results/HEAD/mean.broad_id.profiles.txt')
    dmso_well_profiles = Profiles.load('../results/HEAD/mean.mock.well.profiles.txt')
    active_mask = compute_active_mask(profiles, dmso_well_profiles, percentile)
    profiles.data = profiles.data[active_mask, :]
    profiles._keys = [k for k, active in zip(profiles._keys, active_mask) 
                      if active]
    return profiles

if __name__ == '__main__':

    output_stem, percentile_string = sys.argv[1:]
    percentile = int(percentile_string) / 100.0

    load_properties()

    profiles = load_profiles(percentile)

    broad_ids = [k[0][:13] for k in profiles.keys()]

    unslimmed = read_annotations(broad_ids)
    annotations = slim_annotations(unslimmed)
    all_terms = list(set(t for l in annotations.values() for t in l))

    print len(all_terms), 'terms after slimming'
    print len(annotations), 'of', len(profiles.keys()), 'compounds have annotations'

    smF = np.zeros((len(broad_ids), len(all_terms)), dtype=bool)
    for broad_id, terms in annotations.items():
        i = broad_ids.index(broad_id)
        for term in terms:
            j = all_terms.index(term)
            smF[i, j] = True
    annotated_mask = smF.sum(1) > 0
    profiles.data = profiles.data[annotated_mask, :]
    profiles._keys = [k for k, annotated in zip(profiles._keys, annotated_mask) if annotated]
    smF = smF[annotated_mask, :]

    # Append a compound profile for DMSO.
    dmso_well_profiles = Profiles.load('../results/HEAD/mean.mock.well.profiles.txt')
    profiles.data = np.vstack([profiles.data, [np.mean(dmso_well_profiles.data, 0)]])
    profiles._keys = profiles._keys + [('DMSO',)]
    smF = np.vstack([smF, np.zeros((1, smF.shape[1]), dtype=bool)])

    profiles.save(output_stem + '.txt')
    cpa.util.pickle(output_stem + '.pickle', smF, profiles, all_terms)

