#!/usr/bin/env python

import docx
import matplotlib
matplotlib.use('Agg')
import sys
import numpy as np
import pylab
import matplotlib.lines as mpllines
import cpa
import cpa.util
from cpa.profiling.cache import Cache, DummyNormalization
from cpa.profiling.normalization import StdNormalization
import lib
from util import cached

CACHE_DIR = '../../cache/'

cpa.properties.LoadFile("supplement.properties")

@cached("../results/HEAD/plot_compound_effect_magnitude.pickle")
def get_data():
    cache = Cache(CACHE_DIR)
    params = lib.get_standardization_parameters(cache)
    profiles = []
    replicate_profiles = []
    images_per_broad_id = lib.get_compound_image_keys_per_broad_id()
    for broad_id in lib.get_broad_ids_in_order():
        per_plate_profiles = []
        per_plate_means = []
        for plate, image_keys in images_per_broad_id[broad_id].items():
            data = cache.load(image_keys, normalization=DummyNormalization)[0]
            mean, std = params[plate]
            data = (data - mean) / std
            per_plate_profiles.append(data)
            per_plate_means.append(data.mean(0))
        profiles.append(np.vstack(per_plate_profiles).mean(0))
        replicate_profiles.append(np.vstack(per_plate_means))
    return np.array(profiles), replicate_profiles
    
def remove_nan_columns(a):
    mask = np.isnan(a).sum(0) > 0
    return a[:, ~mask]

if __name__ == '__main__':
    compound_profiles, replicate_profiles = get_data()

    mask = np.isnan(compound_profiles).sum(0) > 0
    data = compound_profiles[:, ~mask]
    replicate_profiles = [p[:, ~mask] for p in replicate_profiles]

    pylab.matshow(data.T, cmap=pylab.cm.Greys_r, vmin=-3, vmax=3)
    pylab.xticks([],[])
    pylab.yticks([],[])
    pylab.ylabel('Features')
    pylab.xlabel('Compounds')
    pylab.colorbar()
    fig = matplotlib.pyplot.gcf()
    fig.set_size_inches(6,9)
    pylab.savefig("../results/HEAD/plot_compound_effect_magnitude.pdf")

    h, e = np.histogram(data.ravel(), 100, range=(-10, 10))
    pylab.figure()
    pylab.bar(0.5 * e[:-1] + 0.5 * e[1:], h, e[1] - e[0])
    pylab.xlabel('Feature value standardized by mock-treated cell population')
    pylab.ylabel('Number of compound-feature pairs')
    pylab.savefig("../results/HEAD/plot_compound_effect_magnitude_histogram_ravel.pdf")

    maxima = np.absolute(data).max(0)

    h, e = np.histogram(maxima, 50, range=(0, 25))
    pylab.figure()
    pylab.bar(0.5 * e[:-1] + 0.5 * e[1:], h, e[1] - e[0])
    pylab.xlabel('Maximal feature value across compounds, standardized by mock-treated cell population')
    pylab.ylabel('Number of features')
    pylab.savefig("../results/HEAD/plot_compound_effect_magnitude_histogram_maxima.pdf")

    indices = np.argsort(maxima)
    cache = Cache(CACHE_DIR)
    normalization = DummyNormalization(cache)
    colnames = normalization.colnames
    table = [['Rank', 'Magnitude', 'Feature name']]
    for rank0, i in enumerate(indices):
        table.append(['%d' % (rank0 + 1), '%f' % maxima[i], colnames[i]])
    document = docx.newdocument()
    docbody = document.xpath('/w:document/w:body', namespaces=docx.nsprefixes)[0]
    docbody.append(docx.paragraph('Compound magnitudes', style='b'))
    docbody.append(docx.table(table))
    docx.savedocx(document,
                  docx.coreproperties(title='Compound magnitude table',
                                      subject='',
                                      creator='Vebjorn Ljosa',
                                      keywords=[]),
                  docx.appproperties(), docx.contenttypes(),
                  docx.websettings(),
                  docx.wordrelationships(docx.relationshiplist()),
                  '../results/HEAD/plot_compound_effect_magnitude_ranking.docx')

    variabilities = []
    for replicate in replicate_profiles:
        cv = np.absolute(replicate.std(0) / replicate.mean(0))
        variabilities.append(cv)
    variabilities = np.vstack(variabilities)
    pylab.figure()
    pylab.matshow(variabilities.T, cmap=pylab.cm.Greys_r, vmin=-1, vmax=1)
    pylab.xticks([],[])
    pylab.yticks([],[])
    pylab.ylabel('Features')
    pylab.xlabel('Compounds')
    pylab.colorbar()
    fig = matplotlib.pyplot.gcf()
    fig.set_size_inches(6,9)
    pylab.savefig("../results/HEAD/plot_compound_effect_cv.pdf")


